import os.path
import random

__author__="Mattia Cavenaghi"
__date__ ="$22-set-2010 16.52.55$"

NUMERO_FACCE_PEDINA = 6
NUMERO_PEDINE_TANTRIX = 56
NOME_FILE_IN = "tiles.txt"
NOME_FILE_OUT = "tantrix_domino_py.txt"

class Pedina:
    def __init__(self, indice, rotazioni, facce, visite):
        self.Indice = indice
        self.Rotazioni = rotazioni
        self.Facce = facce
        self.Visite = visite

def getIndiceFacciaSx (indiceDx):
    """
    Dato l'indice della faccia destra di una pedina, la funzione calcola (e ritorna) l'indice della faccia sinistra
    """
    if indiceDx < (NUMERO_FACCE_PEDINA / 2):
        indiceSx = indiceDx + (NUMERO_FACCE_PEDINA / 2)
    else:
        indiceSx = indiceDx - (NUMERO_FACCE_PEDINA / 2)
    return indiceSx

def generaListaInteri (listaPedine):
    """
    Data una lista contenente i dati delle pedine, la funzione crea una lista contenete SOLO gli indici delle pedine
    """
    listaIndPedine = []
    for pedina in listaPedine:
        listaIndPedine.append(pedina.Indice)

    return listaIndPedine

def generaFileSoluzione (nomeFile, listaEstratte, listaDomino):
    """
    Funzione per la generazione del file contenente la soluzione del Tantrix Domino
    """
    fileOut = open(nomeFile, "w")
    separatore = ";"

    fileOut.write("Mattia Cavenaghi" + separatore + "Matricola: 736856\n")
    fileOut.write("Soluzione del Tantrix Domino (versione Python)\n\n")

    fileOut.write("Lista delle pedine disponibili: %s\n" % listaEstratte)
         
    fileOut.write("\nIl Tantrix Domino trovato e': %s\n\n" % generaListaInteri(listaDomino))
    
    fileOut.write("Dettaglio del Tantrix Domino\n")
    fileOut.write("Ind. pedina" + separatore + "Ind. destro" + separatore + "Colore" + separatore + "Ind. sinistro" + separatore + "Colore\n")

    for pedina in listaDomino:
        fileOut.write(str(pedina.Indice) + separatore + str(pedina.Rotazioni) + separatore + str(pedina.Facce[pedina.Rotazioni]))
        fileOut.write(separatore + str(getIndiceFacciaSx(pedina.Rotazioni)) + separatore + str(pedina.Facce[getIndiceFacciaSx(pedina.Rotazioni)]) + "\n")

    fileOut.close()
    
def resettaPedine (listaIn, listaOut, flagVisite):
    """
    Date due liste In ed Out la funzione inizializza a 0 i contatori delle pedine in In e le inserisce in Out
    """
    for pedina in listaIn:
        pedina.Rotazioni = 0
        
        if (flagVisite):
            pedina.Visite = 0
        
        listaOut.append(pedina)
                    
    return listaOut
    

numeroPedineEstratte = input("Quante pedine vuoi usare per il domino? ")

# Lista contenente gli indici delle pedine estratte generate in modo casuale
listaEstratteInt = []

i = 0
while i < numeroPedineEstratte:
    numeroCasuale = random.randint(1, NUMERO_PEDINE_TANTRIX)

    if (not (numeroCasuale in listaEstratteInt)):
        listaEstratteInt.append(numeroCasuale)
        i += 1
        
# Lista contenente: indice, faccia destra e colore delle facce delle pedine estratte
listaEstratte = []

# Controllo ed apertura del file tiles.txt per riempimento della lista contenente le pedine estratte
if not(os.path.isfile(os.path.abspath('') + "\\%s" % NOME_FILE_IN)):
    print "\nIl file \"%s\" non esiste!" % NOME_FILE_IN
    exit(-1)

fileIn = open(NOME_FILE_IN, 'r')
linee = fileIn.readlines()

for i in listaEstratteInt:

    # Per estrarre i colori delle facce estraggo dalla linea acquisita dal file, la sottostringa compresa tra \t e \n
    # inoltre rimuovo gli spazi bianchi
    faccePedina = linee[i][linee[i].find('\t')+1:linee[i].find('\n')].strip().split(' ')

    # La pedina conterra' oltre ai colori delle facce anche l'indice della faccia destra corrente ed il numero di visite per permutazione (inizializzate a 0)
    listaEstratte.append(Pedina(i, 0, faccePedina, 0))

fileIn.close()

# Lista delle pedine domino (la soluzione)
listaDomino = []

# Inserisco la prima pedina estratta disponibile, qualunque essa sia poiche' va sempre bene
listaDomino.append(listaEstratte.pop(0))

listaScartate = []

# Inzio dell'algoritmo: si ferma solo se trova una soluzione VALIDA (vedi descrizione algortimo)
while len(listaDomino) < numeroPedineEstratte:
    # Se il numero di pedine visitate e' >= al numero di pedine disponibili:
    # - rompo la regola aurea tra l'ultima e la penultima pedina del domino (ruoto l'ultima di una posizione)
    # - memorizzo come pedDx l'ultima pedina del domino (forzo la regola aurea al passo successivo)
    # - cancello la pedina ruotata dal domino
    if len(listaEstratte) == 0:
        if len(listaDomino) > 1:
            pedinaSinistra = listaDomino[len(listaDomino) - 2]

            pedinaDestra = listaDomino.pop(len(listaDomino) - 1)
            pedinaDestra.Rotazioni += 1

            if pedinaDestra.Visite < 2:
                listaEstratte = resettaPedine(listaScartate, listaEstratte, False)
                listaScartate = []
        else:
            pedinaSinistra = listaDomino[0]
            pedinaSinistra.Rotazioni += 1
            
            listaDomino = []
            listaDomino.append(pedinaSinistra)
                
            listaEstratte = resettaPedine(listaScartate, listaEstratte, True)

            listaScartate = []
            
            pedinaDestra = listaEstratte.pop(0)

    else:
        # Calcolo le nuove pedine Dx ed Sx
        pedinaSinistra = listaDomino[len(listaDomino) - 1]
        pedinaDestra = listaEstratte.pop(0)

    indiceFacciaDx = pedinaDestra.Rotazioni

    # Controllo della regola aurea (vedi descrizione algoritmo)
    # Faccia destra della pedina sinistra
    facciaDx = pedinaSinistra.Facce[pedinaSinistra.Rotazioni]
    facciaSx = pedinaDestra.Facce[getIndiceFacciaSx(indiceFacciaDx)]

    # In base all'indice della faccia destra corrente, calcolo l'indice della faccia sinistra corrente
    while (indiceFacciaDx < NUMERO_FACCE_PEDINA) & (facciaSx != facciaDx):
        indiceFacciaDx += 1
        facciaSx = pedinaDestra.Facce[getIndiceFacciaSx(indiceFacciaDx)]
        
    pedinaDestra.Visite += 1
    
    if indiceFacciaDx < NUMERO_FACCE_PEDINA:
        if facciaSx == facciaDx:
            pedinaDestra.Rotazioni = indiceFacciaDx

            listaDomino.append(pedinaDestra)

            if len(listaScartate) > 0:
                listaEstratte = resettaPedine(listaScartate, listaEstratte, False)
                listaScartate = []
    else:
        listaScartate.append(pedinaDestra)

print "\nLe pedine estratte sono: %s" % listaEstratteInt

print "La soluzione e': %s" % generaListaInteri(listaDomino)

generaFileSoluzione(NOME_FILE_OUT, listaEstratteInt, listaDomino)

print "Per i dettagli consultare il file dei risultati: \"%s\"" % NOME_FILE_OUT
