﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;

namespace TantrixDomino
{
    class Program
    {
        // Definizione delle costanti
        public const int NUMERO_FACCE_PEDINA = 6;
        public const int NUMERO_PEDINE_TANTRIX = 56;
        public const string FILE_OUT = @"tantrix_domino_cs.txt";

        /// <summary>
        /// Oggetto che identifica l'enitià pedina
        /// </summary>
        public class Pedina
        {
            public Pedina()
            { }

            public Pedina(int indice, int rotazioni, List<char> facce)
            {
                this.Indice = indice;
                this.Rotazioni = rotazioni;
                this.Facce = facce;
            }

            private int indice = 0;

            public int Indice
            {
                get { return indice; }
                set { indice = value; }
            }

            private int rotazioni = 0;

            public int Rotazioni
            {
                get { return rotazioni; }
                set { rotazioni = value; }
            }

            private List<char> facce = new List<char>(NUMERO_FACCE_PEDINA);

            public List<char> Facce
            {
                get { return facce; }
                set { facce = value; }
            }

            public char FacciaSinistra()
            {
                if (Rotazioni < (NUMERO_FACCE_PEDINA / 2))
                    return Facce.ElementAt(Rotazioni + (NUMERO_FACCE_PEDINA / 2));
                else
                    return Facce.ElementAt(Rotazioni - (NUMERO_FACCE_PEDINA / 2));
            }

            public char FacciaDestra()
            {
                return Facce.ElementAt(Rotazioni);
            }

            private int visite = 0;

            public int Visite
            {
                get { return visite; }
                set { visite = value; }
            }
        }

        /// <summary>
        /// Dato l'indice della faccia destra della pedina calcola l'indice della faccia sinistra
        /// </summary>
        /// <param name="indFaxDx">Indice della faccia destra</param>
        /// <returns>Indice della faccia sinistra</returns>
        public static int getIndiceFaxSx(int indFaxDx)
        {
            if (indFaxDx < (NUMERO_FACCE_PEDINA / 2))
                return indFaxDx + (NUMERO_FACCE_PEDINA / 2);
            else
                return indFaxDx - (NUMERO_FACCE_PEDINA / 2);
        }

        /// <summary>
        /// Date due liste di pedine In e Out, re-inizializzo a 0 i contatori delle pedine In inserendole in Out
        /// </summary>
        /// <param name="listaIn">Lista delle pedine da resettare (scartate)</param>
        /// <param name="listaOut">Lista delle pedine resettate (estratte)</param>
        /// <param name="flagVisite">Flag per l'inizializzazione del contatore visite</param>
        /// <returns>Lista con pedine resettate</returns>
        public static List<Pedina> resetPedine(List<Pedina> listaIn, List<Pedina> listaOut, bool flagVisite)
        {
            foreach (Pedina pedina in listaIn)
            {
                pedina.Rotazioni = 0;

                if (flagVisite)
                    pedina.Visite = 0;

                listaOut.Add(pedina);
            }

            return listaOut;
        }

        public static void Main(string[] args)
        {
            #region Generazione (casuale) della lista contenente gli indici delle pedine da usare
            
            Random randomizzatore = new Random();

            int numeroPedineEstratte = 0;

            // Piccolo controllo sul valore inserito dall'utente (sul range)...
            while (numeroPedineEstratte > NUMERO_PEDINE_TANTRIX || numeroPedineEstratte <= 0)
            {
                Console.Write("Quante pedine vuoi usare per il domino? ");

                // ... e sul tipo di dato passato (solo interi)
                int.TryParse(Console.ReadLine(), out numeroPedineEstratte);
            }

            int[] listaEstratteInt = new int[numeroPedineEstratte];
            string listaEstratteStr = string.Empty;

            int i = 0;
            while (i < numeroPedineEstratte)
            {
                int numeroCasuale = randomizzatore.Next(1, NUMERO_PEDINE_TANTRIX + 1);

                if (!listaEstratteInt.Contains(numeroCasuale))
                {
                    listaEstratteInt[i] = numeroCasuale;
                    listaEstratteStr += " " + numeroCasuale.ToString();

                    i++;
                }
            }

            //// Senza backtracking
            //listaEstratteInt = new int[] { 1, 2, 3 };

            //// Backtracking con sostituzione pedina
            //listaEstratteInt = new int[] { 1, 2, 44 };

            //// Backtracking fino a radice
            //listaEstratteInt = new int[] { 1, 6, 28 };

            ////// Backtracking fino a radice e con sostituzione pedina
            //listaEstratteInt = new int[] { 1, 39, 44 };
            //listaEstratteInt = new int[] { 13, 30, 9 };
            //listaEstratteInt = new int[] { 55, 15, 24 };
            //listaEstratteInt = new int[] { 18, 40, 6 };
            //listaEstratteInt = new int[] { 36, 13, 22 };
            //listaEstratteInt = new int[] { 41, 13, 49 };

            //// Regolare...
            //listaEstratteInt = new int[] { 10, 45, 7, 44 };
            //listaEstratteInt = new int[] { 39, 26, 51, 12 };
            
            Console.WriteLine(String.Format("\nLe pedine estratte sono:{0}", listaEstratteStr));

            #endregion Generazione (casuale) della lista contenente gli indici delle pedine da usare

            #region Conversione del file 'tiles.txt' in vettore di stringhe

            string[] setPedineTmp = TantrixDominoCS.Properties.Resources.tiles.Split('\n');
            string[] setPedine = new string[NUMERO_PEDINE_TANTRIX];

            try
            {
                Array.Copy(setPedineTmp, 1, setPedine, 0, NUMERO_PEDINE_TANTRIX);
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.Message);
            }

            #endregion Conversione del file 'tiles.txt' in vettore di stringhe

            #region Creazione della lista di pedine da usare (tiles.txt -> lista[indPedina, indFaxDx, facce])

            List<Pedina> listaEstratte = new List<Pedina>();

            // Dati gli indici delle pedine estratte creo un vettore di oggetti Pedina, secondo l'ordine di estrazione originale
            foreach (int indiceEstratta in listaEstratteInt)
            {
                string[] datiPedina = setPedine[indiceEstratta - 1].Split('\t');

                // Creazione lista delle facce della singola pedina
                List<char> faccePedina = new List<char>();
                foreach (char faccia in datiPedina[1].Replace(" ", "").ToCharArray())
                    faccePedina.Add(faccia);

                // Creazione della pedina ed inserimento nell'insieme di pedine estratte
                listaEstratte.Add(new Pedina(indiceEstratta, 0, faccePedina));
            }

            #endregion Creazione della lista di pedine da usare (tiles.txt -> lista[indPedina, indFaxDx, facce])

            #region Creazione del Tantrix domino

            List<Pedina> listaDomino = new List<Pedina>();
            listaDomino.Add(listaEstratte.ElementAt(0));

            listaEstratte.RemoveAt(0);

            List<Pedina> listaScartate = new List<Pedina>();

            while (listaDomino.Count < numeroPedineEstratte)
            {
                Pedina pedinaSinistra = new Pedina();
                Pedina pedinaDestra = new Pedina();

                if (listaEstratte.Count == 0)
                {
                    if (listaDomino.Count > 1)
                    {
                        // Forzo la rottura del domino ruotando di una posizione l'ultima pedina
                        pedinaSinistra = listaDomino.ElementAt(listaDomino.Count - 2);

                        pedinaDestra = listaDomino.ElementAt(listaDomino.Count - 1);
                        pedinaDestra.Rotazioni++;

                        listaDomino.RemoveAt(listaDomino.Count - 1);

                        if (pedinaDestra.Visite < 2)
                        {
                            listaEstratte = resetPedine(listaScartate, listaEstratte, false);

                            listaScartate = new List<Pedina>();
                        }
                    }
                    else
                    {
                        // Caso non di base: la pedina è la radice
                        pedinaSinistra = listaDomino.ElementAt(0);
                        pedinaSinistra.Rotazioni++;

                        listaDomino = new List<Pedina>();
                        listaDomino.Add(pedinaSinistra);

                        listaEstratte = resetPedine(listaScartate, listaEstratte, false);

                        listaScartate = new List<Pedina>();

                        pedinaDestra = listaEstratte.ElementAt(0);
                        listaEstratte.RemoveAt(0);
                    }
                }
                else
                {
                    pedinaSinistra = listaDomino.ElementAt(listaDomino.Count - 1);
                    pedinaDestra = listaEstratte.ElementAt(0);

                    listaEstratte.RemoveAt(0);
                }

                int indFaxDx = pedinaDestra.Rotazioni;
                char facciaSx = pedinaDestra.FacciaSinistra();

                // Rotazione della pedina dx e comparazione con la pedina sx
                while ((indFaxDx < NUMERO_FACCE_PEDINA) && !(pedinaSinistra.FacciaDestra().Equals(facciaSx)))
                {
                    indFaxDx++;
                    facciaSx = pedinaDestra.Facce.ElementAt(getIndiceFaxSx(indFaxDx));
                }

                pedinaDestra.Visite++;

                // Memorizzazione della pedina destra rotata nella lista domino
                if (indFaxDx < NUMERO_FACCE_PEDINA)
                {
                    if (pedinaSinistra.FacciaDestra().Equals(facciaSx))
                    {
                        // Memorizzo nel domino la pedina ruotata
                        pedinaDestra.Rotazioni = indFaxDx;

                        listaDomino.Add(pedinaDestra);

                        if (listaScartate.Count > 0)
                        {
							listaEstratte = resetPedine(listaScartate, listaEstratte, false);

                            listaScartate = new List<Pedina>();
                        }
                    }
                }
                else
                    listaScartate.Add(pedinaDestra);
            }

            listaEstratteStr = String.Empty;

            foreach (Pedina pedina in listaDomino)
                listaEstratteStr += " " + pedina.Indice;

            Console.WriteLine("\nLa soluzione del Tantrix domino e':{0}", listaEstratteStr);

            #endregion Creazione del Tantrix domino

            #region Creazione del file dei risultati

            Console.WriteLine(String.Format("\nPer i dettagli consultare il file dei risultati: \'{0}\'.", FILE_OUT));

            StreamWriter sw = new StreamWriter(FILE_OUT, false);

            sw.WriteLine("Mattia Cavenaghi;Matricola: 736856\nSoluzione del Tantrix Domino (versione C#)");

            sw.WriteLine(String.Format("\nLista delle pedine disponibili: {0}", listaEstratteStr));
            
            listaEstratteStr = string.Empty;

            foreach (Pedina pedina in listaDomino)
                listaEstratteStr += " " + pedina.Indice;

            sw.WriteLine(String.Format("\nIl Tantrix Domino trovato e': {0}", listaEstratteStr));

            sw.WriteLine("\nDettaglio del Tantrix Domino\nInd. pedina;Ind. destro;Colore;Ind. sinistro;Colore");

            foreach (Pedina pedina in listaDomino)
            {
                int indDx = pedina.Rotazioni;
                int indSx = getIndiceFaxSx(indDx);

                sw.WriteLine("{0};{1};{2};{3};{4}", pedina.Indice, indDx, pedina.FacciaDestra(), indSx, pedina.FacciaSinistra());
            }

            sw.Close();

            #endregion Creazione del file dei risultati
        }
    }
}
