﻿// Mattia Cavenaghi - matricola 736856
#light

open System
open System.IO
open Microsoft.FSharp.Collections

type ref<'T> =  // '
  { mutable value : 'T } // '

// the ref function, ! and := operators look like this:
let (!) (a:ref<_>) = a.value
let (:=) (a:ref<_>) v = a.value <- v
let ref v = { value = v }

// Costanti necessarie per il programma
let NUMERO_FACCE_PEDINA     = 6
let NUMERO_PEDINE_TANTRIX   = 56
let FILE_IN                 = "tiles.txt" 
let FILE_OUT                = "tantrix_domino_fs.txt";

/// Dato l'indice della faccia destra di una pedina, calcola l'indice della faccia sinistra
let getIndiceFaxSx indiceFaxDx =
    if (indiceFaxDx < (NUMERO_FACCE_PEDINA / 2)) then
        indiceFaxDx + (NUMERO_FACCE_PEDINA / 2)
    else
        indiceFaxDx - (NUMERO_FACCE_PEDINA / 2)

/// Definizione dell'oggetto Pedina
type Pedina (indice : int, rotazioni : int, facce : char[], visite : int) =
    member this.Indice = indice
    member this.Rotazioni = rotazioni
    member this.Facce = facce
    member this.Visite = visite
    member this.FacciaSinistra = this.Facce.[getIndiceFaxSx this.Rotazioni]
    member this.FacciaDestra = this.Facce.[this.Rotazioni]

/// Data una lista di pedine rimuovo la pedina posta all'indice i
let rec rimuoviPedina i listaPedine =
    match i, listaPedine with
    | 0, pedina :: pedine -> pedine
    | i, pedina :: pedine -> pedina :: rimuoviPedina (i - 1) pedine
    | i, [] -> failwith "Indice fuori dai limiti della lista!"

/// Funzione per la generazione di una lista casuale di n indici
let generaListaIndici n = 
    let randomizzatore = new Random()
    Seq.initInfinite (fun _ -> randomizzatore.Next(1, NUMERO_PEDINE_TANTRIX + 1))
    |> Seq.distinct 
    |> Seq.take n
    |> Seq.toList

/// Data una lista di pedine, la funzione permette di inizializzare i contatori a 0
let rec resettaPedine listaPedine flagVisite =
    match listaPedine with
    | [] -> []
    | (pedina : Pedina) :: pedine ->
        let pedina =
          if flagVisite then
            new Pedina(pedina.Indice, 0, pedina.Facce, 0)
          else
            new Pedina(pedina.Indice, 0, pedina.Facce, pedina.Visite)
        
        pedina :: (resettaPedine pedine flagVisite)

/// Controllo della regola aurea con creazione della nuova pedina destra rotata e visitata
let rec regolaAurea indFxDx (pedinaSinistra : Pedina) (pedinaDestra : Pedina) = 
    let facciaDestra = pedinaSinistra.FacciaDestra
    let facciaSinistra = pedinaDestra.Facce.[getIndiceFaxSx indFxDx]

    if (indFxDx < NUMERO_FACCE_PEDINA) && (pedinaSinistra.FacciaDestra <> facciaSinistra) then
        regolaAurea (indFxDx + 1) pedinaSinistra pedinaDestra
    else
        new Pedina(pedinaDestra.Indice, indFxDx, pedinaDestra.Facce, pedinaDestra.Visite + 1)

let main() =
    Console.Write("Quante pedine vuoi usare per il domino? ")
    let numeroPedineEstratte = Int32.Parse(Console.ReadLine())
    let listaEstratteInt = generaListaIndici numeroPedineEstratte

    Console.Write("Le pedine estratte sono: ")
    let rec stampaIndiciInteri listaPedine =
        match listaPedine with 
        | [] -> ()
        | indice :: indici -> 
            Console.Write(" " + indice.ToString())

            stampaIndiciInteri indici

    // Test di funzionamento
    //let listaEstratteInt = [1; 39; 44]
    
    stampaIndiciInteri listaEstratteInt
 
    // Estraggo i dati delle pedine estratte
    let setPedine = File.ReadAllLines(FILE_IN)
    
    /// Data la lista di indici creo la lista delle pedine estratte
    let rec creaListaPedine listaIndici =
        match listaIndici with 
        | [] -> []
        | indice :: indici -> 
            let pedinaFull = (setPedine.[indice]).Split('\t')
            let faccePedina = (pedinaFull.[1]).Trim().Replace(" ","").ToCharArray()
            new Pedina(indice, 0, faccePedina, 0) :: creaListaPedine indici

    let mutable listaEstratte = creaListaPedine listaEstratteInt

    let mutable listaDomino = List.empty<Pedina>
    listaDomino <-  listaDomino @ [listaEstratte.Head]
    listaEstratte <- rimuoviPedina 0 listaEstratte

    let mutable listaScartate = List.empty<Pedina>

    while (listaDomino.Length < numeroPedineEstratte) do
        let mutable pedinaSinistra = new Pedina(0, 0, Array.empty, 0)
        let mutable pedinaDestra = new Pedina(0, 0, Array.empty,0)

        if (listaEstratte.Length = 0) then
            if (listaDomino.Length > 1) then
                // Forzo la rottura del domino ruotando di una posizione l'ultima pedina
                pedinaSinistra <- listaDomino.[listaDomino.Length - 2]
  
                pedinaDestra <- new Pedina(listaDomino.[listaDomino.Length - 1].Indice, listaDomino.[listaDomino.Length - 1].Rotazioni + 1, listaDomino.[listaDomino.Length - 1].Facce, listaDomino.[listaDomino.Length - 1].Visite)

                listaDomino <- rimuoviPedina (listaDomino.Length - 1) listaDomino

                if (pedinaDestra.Visite < 2) then
                    listaEstratte <- listaEstratte @ resettaPedine listaScartate false
                    listaScartate <- List.empty<Pedina>

            else
                // Caso non di base: la pedina è la radice
                pedinaSinistra <- new Pedina(listaDomino.Head.Indice, listaDomino.Head.Rotazioni + 1, listaDomino.Head.Facce, listaDomino.Head.Visite)

                listaDomino <- List.empty<Pedina>
                listaDomino <- listaDomino @ [pedinaSinistra]
  
                listaEstratte <- listaEstratte @ resettaPedine listaScartate true

                listaScartate <- List.empty<Pedina>

                pedinaDestra <- listaEstratte.Head
                listaEstratte <- rimuoviPedina 0 listaEstratte
        else
            pedinaSinistra <- listaDomino.[listaDomino.Length - 1]
            pedinaDestra <- listaEstratte.Head
            
            listaEstratte <- rimuoviPedina 0 listaEstratte

        let mutable indFxDx = pedinaDestra.Rotazioni
        pedinaDestra <- regolaAurea indFxDx pedinaSinistra pedinaDestra

        // Memorizzazione della pedina destra: nel domino (regola aurea ok) o nella lista degli scarti
        if (pedinaDestra.Rotazioni < NUMERO_FACCE_PEDINA) then
            if (pedinaSinistra.FacciaDestra = pedinaDestra.FacciaSinistra) then
                listaDomino <- listaDomino @ [pedinaDestra]
  
                if (listaScartate.Length > 0) then
                    listaEstratte <- listaEstratte @ resettaPedine listaScartate false
                    listaScartate <- List.empty<Pedina>

        else
            listaScartate <- listaScartate @ [pedinaDestra]

    
    File.WriteAllText(FILE_OUT, "Mattia Cavenaghi;Matricola: 736856\nSoluzione del Tantrix Domino (versione F#)")
    File.AppendAllText(FILE_OUT, "\n\nLista delle pedine disponibili:")

    let rec stampaIndiciInteri listaPedine =
        match listaPedine with 
        | [] -> ()
        | indice :: indici -> 
            File.AppendAllText(FILE_OUT, " " + indice.ToString())
            stampaIndiciInteri indici

    stampaIndiciInteri listaEstratteInt

    File.AppendAllText(FILE_OUT, "\n\nIl Tantrix Domino trovato e':")

    Console.Write("\nLa soluzione e': ")
    let rec stampaIndiciPedine listaPedine =
        match listaPedine with 
        | [] -> ()
        | pedina :: pedine -> 
            Console.Write(" " + (pedina : Pedina).Indice.ToString())
            File.AppendAllText(FILE_OUT, " " + (pedina : Pedina).Indice.ToString())
            stampaIndiciPedine pedine

    stampaIndiciPedine listaDomino

    File.AppendAllText(FILE_OUT, "\n\nDettaglio del Tantrix Domino")
    File.AppendAllText(FILE_OUT, "\nInd. pedina;Ind. destro;Colore;Ind. sinistro;Colore")

    let rec stampaDettagliPedine listaPedine =
        match listaPedine with 
        | [] -> ()
        | indice :: indici -> 
            File.AppendAllText
                (
                    FILE_OUT,
                    String.Format("\n{0}{5}{1}{5}{2}{5}{3}{5}{4}", (indice : Pedina).Indice.ToString(), (indice : Pedina).Rotazioni.ToString(), (indice : Pedina).FacciaDestra, getIndiceFaxSx (indice : Pedina).Rotazioni, (indice : Pedina).FacciaSinistra, ";")
                )

            stampaDettagliPedine indici

    stampaDettagliPedine listaDomino

    Console.ReadKey(true) |> ignore

main()